#!/usr/bin/python
### rule-based Khmer name romalization
### by Chenchen Ding, @ NICT, 2016/06
### for python 2.x: tested with python 2.6 and 2.7
### for python 3.x: unichr -> chr, using the "python 3.x" lines

import sys, os, codecs
### for python 2.x
sys.stdin  = codecs.getreader('utf-8')(sys.stdin)
sys.stdout = codecs.getwriter('utf-8')(sys.stdout)
sys.stderr = codecs.getwriter('utf-8')(sys.stderr)
### for python 3.x
### sys.stdin  = codecs.getreader('utf-8')(sys.stdin.buffer)
### sys.stdout = codecs.getwriter('utf-8')(sys.stdout.buffer)
### sys.stderr = codecs.getwriter('utf-8')(sys.stderr.buffer)

CON = 	{
	unichr (0x1780) : "K",
	unichr (0x1781) : "KH",
	unichr (0x1782) : "K",
	unichr (0x1783) : "KH",
	unichr (0x1784) : "NG",
	unichr (0x1785) : "CH",
	unichr (0x1786) : "CHH",
	unichr (0x1787) : "CH",
	unichr (0x1788) : "CHH",# no appear
	unichr (0x1789) : "NH",
	unichr (0x178a) : "D",
	unichr (0x178b) : "TH",
	unichr (0x178c) : "D",
	unichr (0x178d) : "TH",	# no appear
	unichr (0x178e) : "N",
	unichr (0x178f) : "T",
	unichr (0x1790) : "TH",
	unichr (0x1791) : "T",
	unichr (0x1792) : "TH",
	unichr (0x1793) : "N",
	unichr (0x1794) : "B",
	unichr (0x1795) : "PH",
	unichr (0x1796) : "P",
	unichr (0x1797) : "PH",
	unichr (0x1798) : "M",
	unichr (0x1799) : "Y",
	unichr (0x179a) : "R",
	unichr (0x179b) : "L",
	unichr (0x179c) : "V",
	unichr (0x179d) : "SH",	# no use
	unichr (0x179e) : "SS",	# no use
	unichr (0x179f) : "S",
	unichr (0x17a0) : "H",
	unichr (0x17a1) : "L",
	unichr (0x17a2) : "@",
	}
IND = 	{
	unichr (0x17a5) : "I",
	unichr (0x17a6) : "I",	# ?
	unichr (0x17a7) : "OU",
	unichr (0x17a8) : "OUK",# ?
	unichr (0x17a9) : "OU",	# ?
	unichr (0x17aa) : "OR",
	unichr (0x17ab) : "RI",
	unichr (0x17ac) : "RI",	# ?
	unichr (0x17ad) : "LI",	# ?
	unichr (0x17ae) : "LI",	# ?
	unichr (0x17af) : "E",
	unichr (0x17b0) : "AI",	# ?
	unichr (0x17b1) : "OR",
	unichr (0x17b2) : "OR",
	unichr (0x17b3) : "OR",
	}
VOW =	{
	"@"		: {"A" : "A",	"O" : "OR"},	# dummy vowel
	unichr (0x17b6) : {"A" : "A",	"O" : "EA"},
	unichr (0x17b7) : {"A" : "E",	"O" : "I"},
	unichr (0x17b8) : {"A" : "EI",	"O" : "I"},
	unichr (0x17b9) : {"A" : "OE",	"O" : "OE"},
	unichr (0x17ba) : {"A" : "EU",	"O" : "EU"},
	unichr (0x17bb) : {"A" : "O",	"O" : "U"},
	unichr (0x17bc) : {"A" : "O",	"O" : "OU"},
	unichr (0x17bd) : {"A" : "UO",	"O" : "UO"},
	unichr (0x17be) : {"A" : "OEU", "O" : "EU"},
	unichr (0x17bf) : {"A" : "OEUR","O" : "OEUR"},
	unichr (0x17c0) : {"A" : "IE",	"O" : "IE"},
	unichr (0x17c1) : {"A" : "E",	"O" : "E"},
	unichr (0x17c2) : {"A" : "E",	"O" : "E"},
	unichr (0x17c3) : {"A" : "AI",	"O" : "AI"},
	unichr (0x17c4) : {"A" : "OR",	"O" : "OR"},
	unichr (0x17c5) : {"A" : "AO",	"O" : "EOU"},
	unichr (0x17c6) : {"A" : "AM",	"O" : "AM"},	# anusvara
	unichr (0x17c7) : {"A" : "AS",	"O" : "AS"},	# visarga
	unichr (0x17c8) : {"A" : "AK",	"O" : "EAK"},	# yuukaleapintu
	unichr (0x17d0) : {"A" : "OA",	"O" : "OA"},	# samyok sannya
	}
A = 	[
	unichr (0x1780),
	unichr (0x1781),
	unichr (0x1785),
	unichr (0x1786),
	unichr (0x178a),
	unichr (0x178b),
	unichr (0x178e),
	unichr (0x178f),
	unichr (0x1790),
	unichr (0x1794),
	unichr (0x1795),
	unichr (0x179d),
	unichr (0x179e), # a...
	unichr (0x179f),
	unichr (0x17a0),
	unichr (0x17a1),
	unichr (0x17a2),
	unichr (0x17c9), #shifter
	]
O = 	[
	unichr (0x1782),
	unichr (0x1783),
	unichr (0x1784),
	unichr (0x1787),
	unichr (0x1788),
	unichr (0x1789),
	unichr (0x178c),
	unichr (0x178d),
	unichr (0x1791),
	unichr (0x1792),
	unichr (0x1793),
	unichr (0x1796),
	unichr (0x1797),
	unichr (0x1798),
	unichr (0x1799),
	unichr (0x179a),
	unichr (0x179b),
	unichr (0x179c),
	#unichr (0x179e),
	unichr (0x17ca), #shifter
	]

A, O, DEP = set (A), set (O), set ([unichr (0x17b6+x) for x in range (29)])
MUTING	= set ([unichr (x) for x in [0x17cc,0x17cd,0x17cf,0x17d1]])
ANUVIS	= set ([unichr (0x17c6),unichr (0x17c7)])
SHFTRS	= set ([unichr (0x17c9),unichr (0x17ca)])
HCOENG	= unichr (0x17a0)+unichr (0x17d2)
B_TO_P	= unichr (0x1794)+unichr (0x17c9)
BANTOC	= unichr (0x17cb)
COENG	= unichr (0x17d2)

def clean (k) :	return " ".join (k.strip ().split ()) \
		.replace (unichr (0xfeff), "").replace (unichr (0xfffe), "") \
		.replace (unichr (0x17a3), unichr (0x17a2)) \
		.replace (unichr (0x17a4), unichr (0x17a2)+unichr (0x7b6))

def a_dummy_v (k, i) : return i+1 < len (k) and k [i] in CON and not k [i+1] in DEP
def khmerchar (k) :
	k, l = k+" ", range (len (k))
	return "".join ([k [i]+"@" if a_dummy_v (k, i) else k [i] for i in l])

def argstr (c, i) : return i-1 >= 0 and c [i-1] in A
def orgstr (c, i) : return i-1 >= 0 and c [i-1] in O
def rgstrs (c, i) : return "O" if orgstr (c, i) else "A"	# OK as well
def rgstrs (c, i) : return "A" if argstr (c, i) else "O"	# OK as well

def muting (c, i) : return i+1 < len (c) and c [i+1] in MUTING
def bantoc (c, i) : return i+1 < len (c) and c [i+1] == BANTOC
def hcoeng (c, i) : return i-2 >= 0 and c [i-2:i] == HCOENG
def xcoeng (c, i) : return i-2 >= 0 and c [i-1] == COENG
def b_to_p (c, i) : return i+1 < len (c) and c [i:i+2] == B_TO_P

def btwn_c (c, i) : return 0 <= i-1 < i+1 < len (c) and c [i-1] in CON and c [i+1] in CON
def real_v (c, i) : return 0 <= i < len (c) and c [i] != "@" and c [i] in VOW
def vc__cv (c, i) : return btwn_c (c, i) and real_v (c, i-2) and real_v (c, i+2)
def no_vow (c, i) : return not btwn_c (c, i) or vc__cv (c, i)
def befr_v (c, i) : return i+1 < len (c) and c [i+1] in VOW
def aftr_v (c, i) : return i-1 >= 0 and c [i-1] in VOW
def ending (c, i) : return i == len (c)-1 or (i+1 < len (c) and c [i+1] == " ")

def basic (c, r) : 
	for i in range (len (c)) :
		if   c [i] in CON : r [i] = CON [c [i]]
		elif c [i] in IND : r [i] = IND [c [i]]
		elif c [i] in VOW : r [i] = VOW [c [i]][rgstrs (c, i)]
		elif c [i] != " " : r [i] = "@"
		else 		  : r [i] = "_"
	return r

def general (c, r) :
	r = basic (c, r)
	### consonats
	for i in range (len (c)) :
		if   muting (c, i) : r [i] = "@"
		elif bantoc (c, i) : r [i] = (r [i][0]+r [i])[-max (2,len (r [i])):]
		elif b_to_p (c, i) : r [i], r [i+1] = "P", "OR"
		elif hcoeng (c, i) :
			r [i-2] = "@"
			if   r [i] == "K" : r [i] = "G"
			elif r [i] == "S" : r [i] = "Z"
			elif r [i] == "V" : r [i] = "F"
	### vowels 
	for i in range (len (c)) :
		if   c [i] == "@" and no_vow (c, i) : r [i] = "@"
		elif c [i] in ANUVIS and aftr_v (c, i) : r [i] = r [i][1:]
		elif c [i] in SHFTRS :
			if   btwn_c (c, i) : r [i] = "OR"
			elif befr_v (c, i) : r [i] = "@"
	return r

def empirical (c, r) :
	r = general (c, r)
	for i in range (len (c)) :
		### sequential consonants
		if bantoc (c, i) and r [i] == "SS" : r [i] = "S"
		if i+2 < len (c) and r [i+1] == "@" and r [i][-2:] == r [i+2]*2 : r [i] = r [i][:-1]
		if i+1 < len (c) and (r [i], c [i], c [i+1]) == ("KH",unichr (0x1781),"@") : r [i] = "K"
		if i+1 < len (c) and (r [i], c [i], r [i+1]) == ("M",unichr (0x17c6),"NG") : r [i] = "@"
		### stacked consonants
		if xcoeng (c, i) : 
			if r [i] in ["K","NH","D","T","Y","S"]	: r [i] = "@"
			if r [i] == r [i-2]+"H"	: r [i-2], r [i] = r [i], "@"
			if (r [i-2], r [i]) == ("B" ,"R")	: r [i-2], r [i] = "P", "R"
			if (r [i-2], r [i]) == ("B" ,"L")	: r [i-2], r [i] = "P", "L"
			if (r [i-2], r [i]) == ("D" ,"TH")	: r [i-2], r [i] = "TH", "@"
			if (r [i-2], r [i]) == ("T" ,"N")	: r [i-2], r [i] = "TH", "@"
			if ending (c, i+1) and (r [i], r [i+1]) == ("R", "@") : r [i] = "@"
		### R+vowel
		if i-1 >= 0 and (r [i-1], c [i]) == ("R",unichr (0x17b8)) : r [i] = "EY"
		if i-1 >= 0 and (r [i-1], c [i]) == ("R",unichr (0x17be)) : r [i] = "OEU"
		if i-1 >= 0 and (r [i-1], r [i]) == ("R","EY") and ending (c, i) : r [i] = "Y"
		### ending-Y
		if r [i][-1] == "I" and ending (c, i) : r [i] = r [i][:-1]+"Y"
		if i-1 >= 0 and (r [i-1], r [i]) == ("Y", "Y") : r [i] = "I"
		### shortened-EA
		if r [i] == "EA" and bantoc (c, i+1)		: r [i] = "A"
		if r [i] == "EA" and ending (c, i) and i-1 >= 0 :
			if r [i-1] in ["N","L","R","Y","V"]	: r [i] = "A"
			if r [i-1] == "@" and c [i-1] in CON	: r [i] = "A"
	return r

def khmeroma (k, level) : 
	c = khmerchar (clean (k))
	r = [" " if x == " " else "?" for x in c]
	if level == "basic" : return "\t".join (basic (c, r))
	if level == "general" : return "\t".join (general (c, r))
	if level == "empirical" : return "\t".join (empirical (c, r))
	return "\t".join (empirical (c, r))

def main () : 
	try : 
		level = "" if len (sys.argv) < 2 else sys.argv [1]
		for k in sys.stdin : sys.stdout.write (khmeroma (k, level) + "\n")
	except : pass

if __name__ == "__main__" : main ()
